/*
program for listing images as thumbnails and subsequently rotating, flipping, (re)moving them.
intended use is selecting and fixing up photo's coming from a digital camera.

todo:
- fallback to converting to thumbnail on our own if image doesn't have one in exif data.
- make menu's:  rotate 90, 180, 270 (working on the original file), flip by x or y.  remove.
- when rotating/flipping, make sure we keep the exif header and rotate the thumbnail as well?
- figure out how to get the correct width/height values after a resize, in eresized
*/

#include <u.h>
#include <libc.h>
#include <draw.h>
#include <event.h>

typedef struct Thumb Thumb;

struct Thumb {
	char *file;
	Image *i;
	Rectangle r;
	Point ip;
	Point sp;
	Thumb *next;
};

Image *bgimage;
Image *black;
char **files;
int nfiles;
int width, height;
Thumb *first = nil;
Thumb *last = nil;
Font *font;
char *fontname = "/lib/font/bit/fixed/unicode.7x14.font";


enum  {
	Xpad = 10,
	Ypad = 10,
};


char *buttons[] = {
	"start",
	"stop",
	"clear",
	"slower",
	"faster",
	"smaller",
	"larger",
	"write",
	"exit",
	nil,
};

Menu menu = {
	buttons
};


void *
emalloc(int s)
{
	void *p;
	p = malloc(s);
	if(!p)
		sysfatal("malloc");
	return p;
}

int
imgwidth(Image *i)
{
	return i->r.max.x - i->r.min.x;
}

int
imgheight(Image *i)
{
	return i->r.max.y - i->r.min.y;
}


int
exifget(char *file)
{
	int fd[2];
	int ffd;

	if(pipe(fd) < 0)
		sysfatal("pipe: %r");

	switch(rfork(RFNOWAIT|RFFDG|RFREND|RFPROC)){
	case -1:
		sysfatal("rfork: %r");

	case 0:
		fprint(2, "first child okay pid=%d\n", getpid());
		close(fd[0]);

		ffd = open(file, OREAD);
		if(ffd < 0)
			sysfatal("opening %s: %r", file);
		dup(ffd, 0);
		close(ffd);
		dup(fd[1], 1);
		close(fd[1]);
		execl("/bin/exifimage", "exifimage", nil);
		sysfatal("execl: %r");

	default:
		close(fd[1]);
		fprint(2, "see parent okay pid=%d\n", getpid());
		/* fd[0] should be ready to read image */
		break;
	}
	return fd[0];
}


void
thumbdraw(Thumb *t)
{
	draw(screen, rectaddpt(screen->r, t->ip), t->i, nil, ZP);
	string(screen, addpt(screen->r.min, t->sp), black, ZP, font, t->file);
}


void
thumbinsert(Thumb *t)
{
	if(first){
		last->next = t;
		last = t;
	}else
		first = last = t;
}


void
thumbrender(void)
{
	Thumb *ofirst = first;
	Thumb *t;

	first = last = nil;
	for(t = ofirst; t; t = t->next){
		thumbinsert(t);
		thumbdraw(t);
	}
}


Point
thumbwhere(Thumb *t)
{
	int twidth;
	Point p;

	twidth = stringwidth(font, t->file);
	if(imgwidth(t->i) > twidth)
		twidth = imgwidth(t->i);

	if(last){
		if(last->r.max.x + twidth >= width){
			p = Pt(0, last->r.max.y);
		}else{
			p = Pt(last->r.max.x, last->r.min.y);
		}
	}else{
		p = Pt(0, 0);
	}
	t->r = Rect(0, 0, 2*Xpad + twidth, 3*Ypad + font->height + imgheight(t->i));
	t->r.min = addpt(t->r.min, p);
	t->r.max = addpt(t->r.max, p);
	t->ip = addpt(t->r.min, Pt(Xpad + (twidth - imgwidth(t->i)) / 2, Ypad));
	t->sp = addpt(t->r.min, Pt(Xpad + (twidth - stringwidth(font, t->file)) / 2, 2*Ypad+imgheight(t->i)));

	if(last){
		fprint(2, "where: last: %s %d,%d\n", last->file, last->r.min.x, last->r.min.y);
	}
	fprint(2, "where: %d,%d,%d,%d (width=%d, height=%d)\n", t->r.min.x, t->r.min.y, t->r.max.x, t->r.max.y, width, height);
	return p;
}

void
thumbadd(char *file)
{
	Thumb *new;
	int thumbfd;
	fprint(2, "thumbadd file=%s pid=%d\n", file, getpid());

	new = emalloc(sizeof new[0]);
	new->file = file;

	thumbfd = exifget(file);
	if(thumbfd < 0)
		sysfatal("exifget %s: %r", file);

	fprint(2, "reading image...\n");
	new->i = readimage(display, thumbfd, 0);
	fprint(2, "image has been read\n");
	close(thumbfd);
	if(!new->i){
		sysfatal("error reading image: %r");
		/* XXX should resize image by ourselves here... */
	}
	thumbwhere(new);
	new->next = nil;

	thumbinsert(new);
	thumbdraw(new);	
}


void
eresized(int)
{
	int nwidth, nheight;

	fprint(2, "resized\n");

	if(getwindow(display, Refnone) < 0)
		sysfatal("getwindow after resize");

	nwidth = screen->r.max.x - screen->r.min.x;
	nheight = screen->r.max.y - screen->r.min.y;
	if(nwidth != width || nheight != height){
		fprint(2, "new width or height\n");
		width = nwidth;
		height = nheight;

		draw(screen, screen->r, bgimage, nil, ZP);
		// XXX draw all 
	}

	thumbrender();
}



void
usage(void)
{
	fprint(2, "usage: see\n");
	exits("usage");
}

void
main(int argc, char *argv[])
{
	int key;
	Event e;

	ARGBEGIN{
	default:
		usage();
	}ARGEND

	if(initdraw(nil, nil, "see") < 0)
		sysfatal("initdraw failed: %r");

	einit(Emouse|Ekeyboard);

	font = openfont(display, fontname);
	if(!font)
		sysfatal("openfont: %r");

	bgimage = allocimagemix(display, DBlack, DWhite);
	black = allocimagemix(display, DBlack, DBlack);

	width = screen->r.max.x - screen->r.min.x;
	height = screen->r.max.y - screen->r.min.y;
	fprint(2, "width=%d height=%d\n", width, height);

	while(argc--)
		thumbadd(*argv++);

//	newstatus = -1;
	for(;;){
		key = event(&e);

/*
		if(key == Emouse){
			m = e.mouse;
			if(~m.buttons&1){
				newstatus = -1;
			}
			if(m.buttons&1){
				int row, col;
				Point xy;

				xy = subpt(m.xy, screen->r.min);
				row = xy.y / (gridsize+1);
				col = xy.x / (gridsize+1);
				if(row >= 0 && col >= 0 && row < b->nrows && col < b->ncols){
					if(newstatus == -1)
						newstatus = 1 - b->table[row][col];
					b->table[row][col] = newstatus;
				}
				drawlife(1);
			}
			if(m.buttons&4){
				switch(emenuhit(3, &m, &menu)){
				case 1:
					exits(nil);
				}
			}
		}
*/

		if(key == Ekeyboard){
			switch(e.kbdc){
			case 0x04:	/* ctrl-d */
			case 127:		/* del */
			case 'q':
				exits(nil);
			}
		}

	}
}
